{$Stackframes on}
{$define HandleRunErrors }
{ Undefine this for default run error handling }

unit DJMDebug;

interface

uses Windows, SysUtils, Classes, Graphics, Forms, Controls, StdCtrls,
  Buttons, ExtCtrls;

type
  TStackView = (None, Trace, Dump, FullDump);
  TDebugBox = class(TForm)
    Panel1: TPanel;
    Label3: TLabel;
    Address: TLabel;
    Label2: TLabel;
    SrcLine: TLabel;
    Label4: TLabel;
    Routine: TLabel;
    Memo1: TMemo;
    StackView: TRadioGroup;
    Label6: TLabel;
    Message: TLabel;
    Label1: TLabel;
    Exename: TLabel;
    procedure FormResize(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure StackViewClick(Sender: TObject);
    procedure Memo1DblClick(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
    function GetWordAt(start:integer):string;
    { Extracts a word (delimited by whitespace) from the
      memo }
    function LineStart:integer;
    { Returns the character number at the start of the currently
      selected line }
    function LineStop:integer;
    { Returns the character number after the end of the currently
      selected line }
    function CurrentAddress:pointer;
    { Returns the address currently being pointed to in the
      debug window }
    procedure Snapshot(msg:string);
    { Fill in the fields with a current snapshot of the
      caller, and show }
    procedure ShowException;
    procedure ShowCustom(addr:pointer;skip:integer;msg:string;modal:boolean);
  end;

  TDebugDlg = class(TComponent)
  private
    { Private declarations }
    fcaption : string;
    fstackview : TStackview;
    fmodal : boolean;
    fExceptions,fShowHint : boolean;
  protected
    { Protected declarations }
  public
    { Public declarations }
    DebugBox : TDebugBox;
    constructor Create(owner:TComponent); override;
    procedure Snapshot(msg:string);
    { Fill in the fields with a current snapshot of the
      caller, and show }
    procedure ShowException;
    { Show the current exception }
    procedure HandleException(Sender: TObject;E:Exception);
    procedure SetCaption(value:string);
    procedure SetExceptions(value:boolean);
    procedure SetShowHint(value:boolean);
  published
    { Published declarations }
    property Caption:string read fCaption write SetCaption;
    property Exceptions:boolean read fExceptions write SetExceptions;
    property Modal:boolean read fModal write fModal;
    property Stackview:TStackview read fstackview write fstackview default trace;
    property ShowHint:boolean read fShowHint write SetShowHint;
  end;

procedure Register;

var
  DebugBox : TDebugBox;

implementation

uses
  Dialogs,DebugLow;

{$R *.DFM}

procedure TDebugBox.ShowCustom(addr:pointer;skip:integer;msg:string;modal:boolean);
var
  savecursor : TCursor;
  newlines : TStringList;
begin
  savecursor := Screen.cursor;
  Screen.cursor := crHourGlass;
  try
    Message.Caption := msg;
    ExeName.Caption := DebugLow.ExeName;
    Address.Caption := PtrToHex(addr);
    SrcLine.Caption := DebugLow.SrcLine(addr);
    Routine.Caption := DebugLow.Procname(addr);
    Memo1.Lines.Clear;
    Case TStackView(StackView.ItemIndex) of
      Trace : newlines := StackTrace(skip);
      Dump : newlines := StackDump(skip,100,false);
      FullDump : newlines := StackDump(skip,1000,true);
      else
        newlines := nil;
    end;
    if assigned(newlines) then
    begin
      memo1.lines := newlines;
      newlines.free;
    end;  
    if modal then
      ShowModal
    else
      Show;
  finally
    Screen.cursor := savecursor;
  end;  
end;

procedure TDebugBox.Snapshot(msg:string);
begin
  ShowCustom(CallerAddr,2,msg,false);
end;

procedure TDebugBox.ShowException;
var
  e : TObject;
  msg : string;
begin
  e := ExceptObject;
  if assigned(e) then
    msg := Exception(E).Message
  else
    msg := 'No active exception';
  ShowCustom(ExceptAddr,2,msg,false);
end;

procedure TDebugBox.FormResize(Sender: TObject);
begin
  Panel1.width := Width-27;
  StackView.width := Width-27;
  Memo1.width := Width-27;
  Memo1.height := Height - (382-185);
  if Memo1.height < 40 then
    Memo1.height := 40;
end;

procedure TDebugBox.FormCreate(Sender: TObject);
begin
  DebugLow.Initialize(paramstr(0));
end;

procedure TDebugBox.StackViewClick(Sender: TObject);
begin
  if owner is TDebugDlg then
    TDebugDlg(owner).fstackview := TStackView(StackView.ItemIndex);
end;

function TDebugBox.GetWordAt(start:integer):string;
{ Start is measured from 0, like selstart }
const
  whitespace : set of char = [' ',^I,^M,^J];
var
  text : string;
  stop : integer;
begin
  text := Memo1.lines.text;
  stop := start+1;
  while (start > 0) and not (text[start] in whitespace) do
    dec(start);
  inc(start);
  while (stop < length(text)) and not (text[stop] in whitespace) do
    inc(stop);
  result := copy(text,start,stop-start);
end;

function TDebugBox.LineStart:integer;
{ Pos is measured from 0, like selstart }
const
  linebreaks : set of char = [^M,^J];
var
  text : string;
begin
  text := memo1.lines.text;
  result := memo1.selstart;
  while (result > 0) and not (text[result] in linebreaks) do
    dec(result);
end;

function TDebugBox.LineStop:integer;
{ Pos is measured from 0, like selstart }
const
  linebreaks : set of char = [^M,^J];
var
  text : string;
begin
  text := memo1.lines.text;
  result := memo1.selstart;
  while (result < length(text)) and not (text[result+1] in linebreaks) do
    inc(result);
end;

function TDebugBox.CurrentAddress:pointer;
var
  addr,error : integer;
begin
  val('$'+GetWordAt(Memo1.Selstart),addr,error);
  if (error <> 0) and (linestop-linestart > 12) then
    val('$'+GetWordAt(LineStart+12),addr,error);
  if error = 0 then
    result := pointer(addr)
  else
    result := nil;
end;

procedure TDebugBox.Memo1DblClick(Sender: TObject);
var
  target : pointer;
  saveexcepts : TExceptionEvent;
begin
  saveexcepts := application.OnException;
  try
    try
      target := CurrentAddress;
      if assigned(target) then
        raise Exception.Create('Requested exception at '+IntToHex(integer(target),8)) at target;
    except
      on Exception do
      begin end;
    end;
  finally
    application.OnException := saveexcepts;
  end;  
end;

constructor TDebugDlg.create(owner:TComponent);
begin
  inherited;
  fstackview := trace;
  DebugBox := TDebugBox.Create(Self);
  DebugBox.ShowHint := ShowHint;
end;

procedure TDebugDlg.Snapshot(msg:string);
    { Fill in the fields with a current snapshot of the
      caller, and show }
begin
  DebugBox.StackView.ItemIndex := ord(fStackview);
  DebugBox.ShowCustom(CallerAddr,2,msg,fmodal);
end;

procedure TDebugDlg.ShowException;
{ Show the current exception }
begin
  DebugBox.StackView.ItemIndex := ord(fStackview);
  if ExceptObject <> nil then
    DebugBox.ShowCustom(ExceptAddr,2,Exception(ExceptObject).Message,fmodal)
  else
    DebugBox.ShowCustom(ExceptAddr,2,'No active exception',fmodal);
end;

procedure TDebugDlg.HandleException(Sender:Tobject;E:Exception);
begin
  DebugBox.StackView.ItemIndex := ord(fStackView);
  DebugBox.ShowCustom(ExceptAddr,2,E.Message,fmodal);
end;

procedure TDebugDlg.SetExceptions(value:boolean);
begin
  if fExceptions <> value then
  begin
    fExceptions := value;
    if fExceptions then
      application.OnException := HandleException
    else
      application.OnException := Nil
  end;
end;

procedure TDebugDlg.SetShowHint(value:boolean);
begin
  fShowHint := value;
  DebugBox.ShowHint := value;
end;

procedure TDebugDlg.SetCaption(value:string);
begin
  fCaption := value;
  DebugBox.Caption := value;
end;

procedure Register;
begin
  RegisterComponents('DJM', [TDebugDlg]);
end;

{$ifdef handlerunerrors}
procedure MyExitProc;
begin
  if assigned(erroraddr) then
  begin
    { Initialize just in case }
    DebugLow.Initialize(paramstr(0));
    showmessage(Format('Run time error %d at %.8x in %s at line %s',
                        [ExitCode,integer(ErrorAddr),DebugLow.Procname(Pchar(ErrorAddr)-1),
                                            DebugLow.SrcLine(pchar(ErrorAddr)-1)]));
    erroraddr := nil;
  end;
end;

initialization
  AddExitProc(MyExitProc);
{$endif}

end.
